require(tidyr)
require(svMisc)
require(CBPS)
require(geiger)
require(lsa)
require(DescTools)
require(MVTests)
require(caret)
require(dbarts)
require(parallel)
require(gbm)
require(pbmcapply)
require(plyr)
require(data.table)
RNGkind("L'Ecuyer-CMRG")

require(schoenberg)
require(lsr)
require(car)
require(psych)
require(quickmatch)
require(devtools)
require(Matching)
require(rgenoud)
require(optmatch)
require(Hmisc)
require(rio)
require(sjstats)
require(schoolmath)
require(MatchIt)
require(magrittr)
require(xlsx)
require(writexl)
require(ncvreg)
require(np)
require(dplyr)
require(tidyverse)
require(survey)
require(sampling)
require(KernSmooth)
require(caTools)
require(gridExtra)
require(lars)
require(glmnet)
require(Metrics)
Sample=Vectorize(sample)
Dat = function(x, p, size, seed){
  set.seed(seed)
  return(x[x$part==p,][sample(nrow(x[x$part==p,]), size),])
}
Dat.v = Vectorize(Dat, vectorize.args=c("size", "p"))
require(MASS)
require(mgcv) 
require(parallel)
require(ranger)
require(xgboost)
library(dplyr)
library(mgcv)
library(tidymv)
require(haven)
`%notin%` <- Negate(`%in%`)

require(survey)
require(svrep)
require(svMisc)
require(tidyverse)
require(tidymodels)
require(splines)
require(cpsR)
require(usethis)
require(cpsvote)
require(gmp)
require(pbapply)
require(janitor)

## API key 
usethis::edit_r_environ()


# instead of by person, let's repeat the MMIE experiment for house

vars_list_B = c('H_SEQ', # sequence number for house
              'HSUP_WGT', #  survey weight
              'PROP_TAX', # annual property taxes; apparently not available in 2024
         'HTOTVAL', # Total household income
         'FPOVCUT', # low income cutoff dollar amount
         'HINTVAL', # interest income
         'H_NUMPER', # of people in household
         'HUNDER18', # number of minors per household
         'HSS_YN' # getting SS? (factor)
         )

vars_list_A = c('H_SEQ', # sequence number for house
                'HSUP_WGT', #  survey weight
                'HTOTVAL', # Total household income
                'FPOVCUT', # low income cutoff dollar amount
                'HINTVAL', # interest income
                'H_NUMPER', # of people in household
                'HUNDER18', # number of minors per household
                'HSS_YN' # getting SS? (factor)
)

cps_1998 = get_asec(year = 1998,
                    vars = vars_list_B)

cps_1999 = get_asec(year = 1999,
                    vars = vars_list_B)

cps_2022 = get_asec(year = 2022,
                    vars = vars_list_A)

cps_2024 = get_asec(year = 2024,
                    vars = vars_list_A)

B_real =
  rbind(cps_1998 %>%
    mutate(
           hss_yn = factor(hss_yn)) %>%
    dplyr::select(tolower(vars_list_B)) %>%
    setNames(c('seq', 'weight', 'y', paste0('x', 1:(ncol(.)-3)))) %>%
      distinct(seq, .keep_all = TRUE),
    cps_1999 %>%
      mutate(
        hss_yn = factor(hss_yn)) %>%
      dplyr::select(tolower(vars_list_B)) %>%
      setNames(c('seq', 'weight', 'y', paste0('x', 1:(ncol(.)-3)))) %>%
      distinct(seq, .keep_all = TRUE)
  )

num_repY = B_real %>%
  get_dupes(y) %>% 
  tally() %>%
  as.numeric()


num_repL2 = 
  (1:nrow(B_real)) %>%
  sapply(FUN = function(x){sqrt(sum(B_real[x, c('y',
                                                paste0('x', 1:(p-1)))])^2)}) %>%
  as.data.frame() %>%
  setNames('norm') %>%
  get_dupes(norm) %>%
  tally() %>%
  as.numeric()

A22_real =
  cps_2022 %>%
  mutate(
    hss_yn = factor(hss_yn)) %>%
  dplyr::select(tolower(vars_list_A)) %>%
  setNames(c('seq', 'weight',  paste0('x', 1:6))) %>%
  distinct(seq, .keep_all = TRUE)

A24_real =
  cps_2024 %>%
  mutate(
    hss_yn = factor(hss_yn)) %>%
  dplyr::select(tolower(vars_list_A)) %>%
  setNames(c('seq', 'weight',  paste0('x', 1:6))) %>%
  distinct(seq, .keep_all = TRUE)



realdata = function(A, B, p){

  
  nA = nrow(A)
  nB = nrow(B)
  p = 6
  
  
  # baseline
  N = sum(A$weight)
  #mean.A = sum(A$y * A$weight) / N
  mean.B = mean(B$y)
  
  
  ## NPMIEG
  
  fx_gam_lm = 
  c(paste0('ns(x', 1:(p-1)) %>%
    paste0(', df = 10)'), 'x6') %>%
    paste(collapse = '+') %>%
    reformulate('y')
  
  
  B.gcv = lm(fx_gam_lm,
             data = B) # has nearly identical results with mgcv::gam, and is faster
  
  predlars_A =if( any(class(B.gcv) == "try-error")){NA} else{predict(B.gcv, newdata=A %>% dplyr::select(paste0("x",
                                                                                                               1:p)))}
  predlars_B = if( any(class(B.gcv) == "try-error")){NA} else{predict(B.gcv, newdata=B %>% dplyr::select(paste0("x",
                                                                                                                1:p)))}
  GAM = (1/N)*sum(predlars_A*A$weight)
  
  
  
  # # # DR - GAM - GAM
  
  #  Sampling Setup
  A.r = A %>%
    mutate(y = predlars_A) %>%
    dplyr::select(paste0('x', 1:p), 'y', 'weight')
  
  
  
  
  B.r = B %>%
    dplyr::select(paste0("x", 1:p), y) %>%
    mutate(weight = NA)
  
  A_or_B = union_all(A.r, B.r) %>%
    mutate(samp = rep(c('A','B'), c(nA, nB)))
  
  # Score function 
  t0 = rep(0, times = 5) %>% matrix() # initial
  U.theta = function(A.r, B.r, t.m){
    p = 5
    A.r = A.r %>%
      dplyr::select(paste0('x', 1:p), everything())
    
    B.r = B.r %>%
      dplyr::select(paste0('x', 1:p), everything())
    
    B.xi_trans = as.matrix(B.r[1:nB, 1:p])  
    A.xi_trans = as.matrix(A.r[1:nA, 1:p])
    num_pi = exp(A.xi_trans%*%t.m)
    dem_pi = (1+exp(A.xi_trans%*%t.m))
    hatpi = ifelse(num_pi == Inf & dem_pi == Inf, 1, num_pi/dem_pi)
    
    p1 =  (1:nB) %>%
      sapply(FUN = function(x){
        matrix(B.xi_trans[x, 1:p], 
               nrow = p,
               ncol = 1)
      }, simplify =FALSE)  %>%
      lapply(apply, MARGIN =1, FUN = as.numeric) %>%
      Reduce(x=., '+') %>% t() %>% t()
    p2 =  (1:nA) %>%
      sapply(FUN = function(x){
        (A.r$weight[x]*hatpi[x])* matrix(A.xi_trans[x, 1:p], 
                                         nrow = p,
                                         ncol = 1)
      }
      , simplify = FALSE) %>%
      Reduce(x=., '+')
    
    return(p1 - p2)
  }
  H.theta = function(A.r, B.r, t.m){
    p = 5
    A.r = A.r %>%
      dplyr::select(paste0('x', 1:p), everything())
    
    
    A.xi_trans = as.matrix(A.r[1:nA, 1:p])
    num_pi = exp(A.xi_trans%*%t.m)
    dem_pi = (1+exp(A.xi_trans%*%t.m))
    hatpi = ifelse(num_pi == Inf & dem_pi == Inf, 1, num_pi/dem_pi)
    p2 =  (1:nA) %>%
      sapply(FUN = function(x){(A.r$weight[x]*hatpi[x]*(1-hatpi[x]))*
          matrix(A.xi_trans[x, 1:p], nrow =p, ncol =1) %*%
          matrix(A.xi_trans[x, 1:p], nrow =1, ncol =p)
      }, simplify =FALSE) %>%
      lapply(apply, MARGIN =1, FUN = as.numeric) %>%
      Reduce(x=., '+') %>% t() %>% t() 
    return(p2)
    
  }
  score_func = function(A.r, B.r, t.m){
    return(t.m + solve(H.theta(A.r, B.r, t.m))%*%U.theta(A.r, B.r, t.m))
  }
  
  # Solving Score function
  tol = 1e-06 
  max_iter = 300
  solve_score = function(A.r, B.r, t0, tol, max_iter){
    U.theta = function(A.r, B.r, t.m){
      p=5
      A.r = A.r %>%
        dplyr::select(paste0('x', 1:p), everything())
      
      B.r = B.r %>%
        dplyr::select(paste0('x', 1:p), everything())
      
      B.xi_trans = as.matrix(B.r[1:nB, 1:p])  
      A.xi_trans = as.matrix(A.r[1:nA, 1:p])
      num_pi = exp(A.xi_trans%*%t.m)
      dem_pi = (1+exp(A.xi_trans%*%t.m))
      hatpi = ifelse(num_pi == Inf & dem_pi == Inf, 1, num_pi/dem_pi)
      
      p1 =  (1:nB) %>%
        sapply(FUN = function(x){
          matrix(B.xi_trans[x, 1:p], 
                 nrow = p,
                 ncol = 1)
        }, simplify =FALSE)  %>% 
        lapply(apply, MARGIN =1, FUN = as.numeric) %>%
        Reduce(x=., '+') %>% t() %>% t() 
      p2 =  (1:nA) %>%
        sapply(FUN = function(x){
          (A.r$weight[x]*hatpi[x])* matrix(A.xi_trans[x, 1:p], 
                                           nrow = p,
                                           ncol = 1)
        }
        , simplify = FALSE) %>% 
        lapply(apply, MARGIN =1, FUN = as.numeric) %>%
        Reduce(x=., '+') %>% t() %>% t() 
      
      return(p1 %>% apply(MARGIN = 1, FUN = as.numeric) %>% t() %>% t() -
            p2 %>% apply(MARGIN = 1, FUN = as.numeric) %>% t() %>% t() )
    }
    H.theta = function(A.r, B.r, t.m){
      p = 5
      A.r = A.r %>%
        dplyr::select(paste0('x', 1:p), everything())
      
      
      A.xi_trans = as.matrix(A.r[1:nA, 1:p])
      num_pi = exp(A.xi_trans%*%t.m)
      dem_pi = (1+exp(A.xi_trans%*%t.m))
      hatpi = ifelse(num_pi == Inf & dem_pi == Inf, 1, num_pi/dem_pi)
      p2 =  (1:nA) %>%
        sapply(FUN = function(x){(A.r$weight[x]*hatpi[x]*(1-hatpi[x]))*
            matrix(A.xi_trans[x, 1:p], nrow =p, ncol =1) %*%
            matrix(A.xi_trans[x, 1:p], nrow =1, ncol =p)
        }, simplify =FALSE) %>%
        Reduce(x=., '+') 
      
      return(p2)
      
    }
    score_func = function(A.r, B.r, t.m){
      return(t.m + solve(H.theta(A.r, B.r, t.m))%*%U.theta(A.r, B.r, t.m))
    }
    
    th = rep(list(NULL), max_iter)
    for(i in 1:max_iter){
      if(i == 1){
        th[[i]] = t0 + solve(H.theta(A.r, B.r, t0))%*%U.theta(A.r, B.r, t0)
      } else{
        th[[i]] = th[[i-1]] +
          solve(H.theta(A.r, B.r, th[[i-1]]))%*%U.theta(A.r, B.r, th[[i-1]])
        
        e_dist = dist(rbind(as.numeric(th[[i]]),
                            as.numeric(th[[i-1]])))
        if(e_dist <= tol){
          break
        }
      }
    }
    
    th_final = 
      (th %>% purrr::discard(is.null))[i] %>%
      unlist() %>%
      matrix(nrow = 5, ncol =1)
    
    return(th_final)
  }
  th_final = try(solve_score(A.r,
                             B.r,
                             t0,
                             tol,
                             max_iter), silent = TRUE)
  
  if(any(class(th_final) == 'try-error')){
    prop_mean = NA
    DR_mean = NA
  } else{
    
    B = B %>%
      mutate(est_probs = exp(as.matrix(B.r[, 1:(p-1)])%*%th_final)/
               (1+exp(as.matrix(B.r[, 1:(p-1)])%*%th_final))) %>%
      mutate(est_weights = 1/est_probs)
    
    prop_mean = sum(B$y*B$est_weights)/sum(B$est_weights)
    DR_mean = (1/sum(B$est_weights))*sum(B$est_weights * (B$y - predlars_B)) +
      (1/sum(A$weight))*sum(A$weight*predlars_A) 
    
  }
  
  # Statistical Matching
  
  MMIE_code = function(A, B, b_model, novel.dat = A_or_B, N, method, p){
    
    fx_gam_lm = c(paste0('ns(x', 1:(p-1)) %>%
        paste0(', df = 10)'), 'x6') %>%
      paste(collapse = '+') %>%
      reformulate('y')
    
    
    fx_novel = paste0("x", 1:p) %>%
      paste(collapse = "+") %>% 
      reformulate("samp")
    fx_novel_nn = c(
      paste0("x", 1:p),
      'y') %>%
      paste(collapse = "+") %>% 
      
      reformulate("samp")
    
    
    
    # matching step
    if(method == 'exact'){
      novel.match = try(matchit(fx_novel,
                                method = 'exact',
                                verbose = FALSE,
                                data = novel.dat), silent = TRUE)
      
      if(any(class(novel.match) == 'try-error')){
        novel.df = data.frame(y = NA, x1 = NA, x2 = NA, x3 = NA, x4 = NA, x5 = NA, x6 = NA)
      } else{
        novel.df = match.data(novel.match) %>%
          filter(samp == 'B')   %>%  mutate(A_Prob = nA/N) %>%
          mutate(weight = 1/A_Prob)
      }
      
      
    } else if(method == 'cem'){
      novel.match = matchit(fx_novel,
                            method = 'cem',
                            verbose = FALSE,
                            cutpoints = 'q5',
                            data = novel.dat)
      
      novel.df = match.data(novel.match) %>%
        filter(samp == "B")   %>%  mutate(A_Prob = nA/N) %>%
        mutate(weight = 1/A_Prob)
    } else if(method == 'nn'){
      novel.match = matchit(fx_novel_nn,
                            method = 'nearest',
                            distance = 'mahalanobis',
                            verbose = FALSE,
                            data = novel.dat)
      
      novel.df = match.data(novel.match) %>%
        filter(samp == "B")   %>%  mutate(A_Prob = nA/N) %>%
        mutate(weight = 1/A_Prob)
      
    } else if(method == 'fm_nn'){
      
      A_hat = novel.dat %>%
        filter(samp == 'A')
      
      match_ids = (1:nA) %>%
        sapply(FUN = function(x){which.min(abs(A_hat$y[x]-B$y))},
               simplify = FALSE
        ) %>%
        unlist()
      
      novel.df = B[match_ids, ] %>%
        mutate(weight = N/nA)
      
    } else if(method == 'fm_nn_repY'){
      
      
      A_hat = novel.dat %>%
        filter(samp == 'A')
      
      norm_A = (1:nA) %>%
        sapply(FUN = function(x){sqrt(sum(A_hat[x, c('y',
                                                     paste0('x', 1:(p-1)))])^2)}) 
      norm_B = (1:nB) %>%
        sapply(FUN = function(x){sqrt(sum(B[x, c('y',
                                                 paste0('x', 1:(p-1)))])^2)}) 
      
      match_ids = (1:nA) %>%
        sapply(FUN = function(x){which.min(abs(norm_A[x]-norm_B))},
               simplify = FALSE
        ) %>%
        unlist()
      
      
      novel.df = B[match_ids, ] %>%
        mutate(weight = N/nA)
      
    } 
    
    # HAJ est
    assign(paste0("HAJ_", method),
           sum(novel.df$y * novel.df$weight) / sum(novel.df$weight))
    
    # MMIs
    
    gcvs = try(lm(fx_gam_lm, 
                  data = novel.df),
               silent = TRUE)
    
    if(any(class(gcvs) == 'try-error')){
      assign(paste0("HAJ_", method, "_gcv"), NA)
    } else{
      assign(paste0("HAJ_", method, "_gcv"),
             sum(predict(gcvs, newdata = A)*A$weight) / sum(A$weight))
    }
    
    
    
    
    # FINAL
    
    MMIE_df = data.frame(get(paste0("HAJ_", method)),
                         get(paste0("HAJ_", method, "_gcv")))
    
    
    
    colnames(MMIE_df) = c(paste0('SM', method), 
                          paste0('MMI', method))
    
    
    
    return(MMIE_df)
  }
  
  MMIE2_exact = MMIE_code(A = A,
                          B = B,
                          p = p,
                          N = N,
                          novel.dat = A_or_B,
                          method = 'exact')
  
  MMIE2_cem= MMIE_code(A = A,
                       B = B,
                       p = p,
                       N = N,
                       novel.dat = A_or_B,
                       method = 'cem')
  MMIE2_fm_nn = MMIE_code(A = A,
                          B = B,
                          p = p,
                          N = N,
                          b_model = B.gcv,
                          novel.dat = A_or_B,
                          method = 'fm_nn')
  
  MMIE2_fm_nn_repY = MMIE_code(A = A,
                               B = B,
                               p = p,
                               b_model = B.gcv,
                               N = N,
                               novel.dat = A_or_B,
                               method = 'fm_nn_repY')
  
  
  MMIE = cbind(MMIE2_exact[1:2],
               MMIE2_cem[1:2],
               MMIE2_fm_nn[1:2],
               MMIE2_fm_nn_repY[1:2])
  
  
  final_dat = cbind(
    mean.B,
    GAM,
    prop_mean,
    DR_mean, 
    MMIE) %>%
    data.frame() %>%
    mutate(mu = 1815) %>% 
    dplyr::select(
      mu,
      mean.B,
      GAM,
      prop_mean,
      DR_mean,
      paste0('SM', c('exact', 'cem', 'fm_nn', 'fm_nn_repY')),
      paste0('MMI', c('exact', 'cem', 'fm_nn', 'fm_nn_repY'))) %>%
    mutate(nA = nA,
           nB = nB)
  
  return(final_dat)
}

set.seed(101)
B_list = replicate(500, expr = B_real[
  sample(1:nrow(B_real), nrow(B_real)),
  ], simplify = FALSE) # order effect?

results = pbmcmapply(FUN = realdata,
                     B = B_list,
                     MoreArgs = list('A' = A22_real),
                     mc.cores = 8) # run if cores are available

# results = pbmapply(FUN = realdata,
#                    B = B_list,
#                    MoreArgs = list(A = A22_real))

setwd('/Users/jeremyflood/OneDrive/Documents/Grad School/2024-2025/Spring 2025/MMIE Revisions/Real Data')

results_organized = data.frame(t(results))


# results_organized[is.na(results_organized)] = ''
# 
# results_organized %>% openxlsx::write.xlsx('realdat.xlsx')

final_results = results_organized %>%
  apply(MARGIN =2, FUN = as.numeric) %>%
  as.data.frame() %>%
  dplyr::select(nA, nB, mu, everything()) %>%
  pivot_longer(cols = 4:ncol(.)) %>%
  group_by(name) %>%
  dplyr::summarize(mean = mean(value, na.rm = TRUE),
                   sd = round(sd(value, na.rm = TRUE), digits = 10)) %>%
  mutate(mu = 1815, nB = nrow(B_real), nA = nrow(A22_real), 
         sd = ifelse(sd == 0.000000e+00 | sd <= 1e-8, 0, sd)) %>%
  mutate(d = (mu-mean), rel_err = abs(mean - mu)/mu) %>%
  arrange(abs(d)) %>%
  dplyr::select(mu, nB, nA, everything()) %>%
  filter(name != 'prop_mean') #%>%
  #openxlsx::write.xlsx('clean_realdat.xlsx')

results_organized %>%
  apply(MARGIN =2, FUN = as.numeric) %>%
  as.data.frame() %>%
  dplyr::select(-c(nA, nB)) %>%
  apply(MARGIN =2, FUN = var) %>%
  round(digits = 2)


# calculating unbiased variance estimator
# 
# setwd('/Users/jeremyflood/OneDrive/Documents/Grad School/2024-2025/Spring 2025/MMIE Revisions/Real Data')
# 
# results_organized = openxlsx::read.xlsx('realdat.xlsx')
# 
# nB = nrow(B_real)
# pi_star = 1-(1-(factorialZ(nB)))^(500)
# results_organized = results_organized %>%
#   dplyr::select(nA, nB, mu, everything()) %>%
#   pivot_longer(cols = 4:ncol(.)) %>%
#   mutate(value = as.numeric(value)) %>%
#   mutate(pi_expanded_value = value * (1/pi_star))
